@@ -28,130 +28,34 @@ from ._util import (
 )
 
 
-class BatchJob(WDL.runtime.task_container.TaskContainer):
+class BatchJobBase(WDL.runtime.task_container.TaskContainer):
+    """
+    Abstract base class implementing the AWS Batch backend for miniwdl TaskContainer. Concrete
+    subclasses add configuration specific to the the shared filesystem in use.
+    """
+
     @classmethod
     def global_init(cls, cfg, logger):
-        cls._set_config_defaults(cfg, logger)
+        cls._submit_lock = threading.Lock()
+        cls._last_submit_time = [0.0]
+        cls._init_time = time.time()
+        cls._describer = BatchJobDescriber()
+
         cls._region_name = detect_aws_region(cfg)
         assert (
             cls._region_name
         ), "Failed to detect AWS region; configure AWS CLI or set environment AWS_DEFAULT_REGION"
 
-        # EFS configuration based on:
-        # - [aws] fsap / MINIWDL__AWS__FSAP
-        # - [aws] fs / MINIWDL__AWS__FS
-        # - SageMaker Studio metadata, if applicable
-        cls._fs_id = None
-        cls._fsap_id = None
-        cls._fs_mount = cfg.get("file_io", "root")
-        assert (
-            len(cls._fs_mount) > 1
-        ), "misconfiguration, set [file_io] root / MINIWDL__FILE_IO__ROOT to EFS mount point"
-        if cfg.has_option("aws", "fs"):
-            cls._fs_id = cfg.get("aws", "fs")
-        if cfg.has_option("aws", "fsap"):
-            cls._fsap_id = cfg.get("aws", "fsap")
-            if not cls._fs_id:
-                cls._fs_id = efs_id_from_access_point(cls._region_name, cls._fsap_id)
-        cls._studio_efs_uid = None
-        sagemaker_studio_efs = detect_sagemaker_studio_efs(logger, region_name=cls._region_name)
-        if sagemaker_studio_efs:
-            (
-                studio_efs_id,
-                studio_efs_uid,
-                studio_efs_home,
-                studio_efs_mount,
-            ) = sagemaker_studio_efs
-            assert (
-                not cls._fs_id or cls._fs_id == studio_efs_id
-            ), "Configured EFS ([aws] fs / MINIWDL__AWS__FS, [aws] fsap / MINIWDL__AWS__FSAP) isn't associated with current SageMaker Studio domain EFS"
-            cls._fs_id = studio_efs_id
-            assert cls._fs_mount.rstrip("/") == studio_efs_mount.rstrip("/") or (
-                cls._fs_mount.startswith(studio_efs_mount).rstrip("/") + "/"
-            ), (
-                "misconfiguration, set [file_io] root / MINIWDL__FILE_IO__ROOT to "
-                + studio_efs_mount.rstrip("/")
-            )
-            cls._studio_efs_uid = studio_efs_uid
-            if not cls._fsap_id:
-                cls._fsap_id = detect_studio_fsap(
-                    logger,
-                    studio_efs_id,
-                    studio_efs_uid,
-                    studio_efs_home,
-                    region_name=cls._region_name,
-                )
-                assert (
-                    cls._fsap_id
-                ), "Unable to detect suitable EFS Access Point for use with SageMaker Studio; set [aws] fsap / MINIWDL__AWS__FSAP"
-            # TODO: else sanity-check that FSAP's root directory equals studio_efs_home
-        assert (
-            cls._fs_id
-        ), "Missing EFS configuration ([aws] fs / MINIWDL__AWS__FS or [aws] fsap / MINIWDL__AWS__FSAP)"
-        if not cls._fsap_id:
-            logger.warning(
-                "AWS BatchJob plugin recommends using EFS Access Point to simplify permissions between containers (configure [aws] fsap / MINIWDL__AWS__FSAP to fsap-xxxx)"
-            )
-        logger.debug(
-            _(
-                "AWS BatchJob EFS configuration",
-                fs_id=cls._fs_id,
-                fsap_id=cls._fsap_id,
-                mount=cls._fs_mount,
-            )
-        )
-
         # set AWS Batch job queue
-        if cfg.has_option("aws", "task_queue"):
-            cls._job_queue = cfg.get("aws", "task_queue")
-        elif sagemaker_studio_efs:
-            cls._job_queue = detect_gwfcore_batch_queue(
-                logger, sagemaker_studio_efs[0], region_name=cls._region_name
-            )
-        assert (
-            cls._job_queue
-        ), "Missing AWS Batch job queue configuration ([aws] task_queue / MINIWDL__AWS__TASK_QUEUE)"
+        cls._job_queue = cfg.get("aws", "task_queue", "")
 
         # TODO: query Batch compute environment for resource limits
         cls._resource_limits = {"cpu": 9999, "mem_bytes": 999999999999999}
-        cls._submit_lock = threading.Lock()
-        cls._last_submit_time = [0.0]
-        cls._init_time = time.time()
-        cls._describer = BatchJobDescriber()
-        logger.info(
-            _(
-                "initialized AWS BatchJob plugin",
-                region_name=cls._region_name,
-                job_queue=cls._job_queue,
-                resource_limits=cls._resource_limits,
-            )
-        )
 
-    @classmethod
-    def _set_config_defaults(cls, cfg, logger):
-        # Set defaults in the cfg: ConfigLoader object so that our subsequent cfg.get* ops won't
-        # throw if the user's custom .cfg file, which we don't necessarily control, omits them.
-        #
-        # This repo's miniwdl_aws.cfg also sets these options, and takes precedence in typical
-        # operations using the auto-built Docker image. But consider the case where a user has
-        # edited/written their own .cfg file, and we release a new plugin version with a new
-        # config option. That'd break their existing .cfg file unless the option has a default
-        # of last resort set here.
-        cfg.plugin_defaults(
-            {
-                "aws": {
-                    "job_timeout": 864000,
-                    "describe_period": 1,
-                    "submit_period": 1,
-                    "submit_period_b": 0.0,
-                    "submit_period_c": 0.0,
-                    "boto3_retries": {"max_attempts": 5, "mode": "standard"},
-                    "retry_wait": 20,
-                    "container_sync": False,
-                    "job_tags": dict(),
-                }
-            }
-        )
+        cls._fs_mount = cfg.get("file_io", "root")
+        assert (
+            cls._fs_mount.startswith("/") and cls._fs_mount != "/"
+        ), "misconfiguration, set [file_io] root / MINIWDL__FILE_IO__ROOT to EFS mount point"
 
     @classmethod
     def detect_resource_limits(cls, cfg, logger):
@@ -161,10 +65,10 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
         super().__init__(cfg, run_id, host_dir)
         self._logStreamName = None
         self._inputs_copied = False
-        # We'll direct Batch to mount EFS inside the task container at the same location we have
-        # it mounted ourselves, namely /mnt/efs. Therefore container_dir will be the same as
-        # host_dir (unlike the default Swarm backend, which mounts it at a different virtualized
-        # location)
+        # We expect the Batch job containers to have the shared filesystem mounted at the same
+        # location we, the workflow job, have it mounted ourselves. Therefore container_dir will be
+        # the same as host_dir (unlike the default Swarm backend, which mounts it at a different
+        # virtualized location)
         self.container_dir = self.host_dir
 
     def copy_input_files(self, logger):
@@ -186,7 +90,7 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
         return os.path.join(self.host_dir, "stderr.txt")
 
     def reset(self, logger) -> None:
-        cooldown = self.cfg.get_float("aws", "retry_wait")
+        cooldown = self.cfg.get_float("aws", "retry_wait", 20.0)
         if cooldown > 0.0:
             logger.info(
                 _(
@@ -206,7 +110,9 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
         Run task
         """
         self._observed_states = set()
-        boto3_retries = self.cfg.get_dict("aws", "boto3_retries")
+        boto3_retries = self.cfg.get_dict(
+            "aws", "boto3_retries", {"max_attempts": 5, "mode": "standard"}
+        )
         try:
             aws_batch = boto3.Session().client(  # Session() needed for thread safety
                 "batch",
@@ -214,9 +120,11 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
                 config=botocore.config.Config(retries=boto3_retries),
             )
             with ExitStack() as cleanup:
+                # prepare the task working directory
+                self._prepare_dir(logger, cleanup, command)
                 # submit Batch job (with request throttling)
                 job_id = None
-                submit_period = self.cfg.get_float("aws", "submit_period")
+                submit_period = self.cfg.get_float("aws", "submit_period", 1.0)
                 while True:
                     with self._submit_lock:
                         if terminating():
@@ -225,7 +133,7 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
                             time.time() - self._last_submit_time[0]
                             >= submit_period * self._submit_period_multiplier()
                         ):
-                            job_id = self._submit_batch_job(logger, cleanup, aws_batch, command)
+                            job_id = self._submit_batch_job(logger, cleanup, aws_batch)
                             self._last_submit_time[0] = time.time()
                             break
                     time.sleep(submit_period / 4)
@@ -236,7 +144,32 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
             logger.error(wrapper)
             raise wrapper
 
-    def _submit_batch_job(self, logger, cleanup, aws_batch, command):
+    def _prepare_dir(self, logger, cleanup, command):
+        # Prepare control files. We do NOT use super().touch_mount_point(...) because it fails if
+        # the desired mount point already exists; which it may in our case after a retry (see
+        # self.host_work_dir() override above.)
+        with open(os.path.join(self.host_dir, "command"), "w") as outfile:
+            outfile.write(command)
+        with open(self.host_stdout_txt(), "w"):
+            pass
+        with open(self.host_stderr_txt(), "w"):
+            pass
+
+        if not self._inputs_copied:
+            # Prepare symlinks to the input Files & Directories
+            container_prefix = os.path.join(self.container_dir, "work/_miniwdl_inputs/")
+            link_dirs_made = set()
+            for host_fn, container_fn in self.input_path_map.items():
+                assert container_fn.startswith(container_prefix) and len(container_fn) > len(
+                    container_prefix
+                )
+                link_dn = os.path.dirname(container_fn)
+                if link_dn not in link_dirs_made:
+                    os.makedirs(link_dn)
+                    link_dirs_made.add(link_dn)
+                symlink_force(host_fn, container_fn)
+
+    def _submit_batch_job(self, logger, cleanup, aws_batch):
         """
         Register & submit AWS batch job, leaving a cleanup callback to deregister the transient
         job definition.
@@ -251,7 +184,7 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
         # concurrent RegisterJobDefinition requests
         job_name = randomize_job_name(job_name)
 
-        container_properties = self._prepare_container_properties(logger, command)
+        container_properties = self._prepare_container_properties(logger)
         job_def = aws_batch.register_job_definition(
             jobDefinitionName=job_name,
             type="container",
@@ -268,7 +201,7 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
 
         self._cleanup_job_definition(logger, cleanup, aws_batch, job_def_handle)
 
-        job_tags = self.cfg.get_dict("aws", "job_tags")
+        job_tags = self.cfg.get_dict("aws", "job_tags", {})
         if "AWS_BATCH_JOB_ID" in os.environ:
             # If we find ourselves running inside an AWS Batch job, tag the new job identifying
             # ourself as the "parent" job.
@@ -278,7 +211,7 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
             jobName=job_name,
             jobQueue=self._job_queue,
             jobDefinition=job_def_handle,
-            timeout={"attemptDurationSeconds": self.cfg.get_int("aws", "job_timeout")},
+            timeout={"attemptDurationSeconds": self.cfg.get_int("aws", "job_timeout", 86400)},
             tags=job_tags,
         )
         logger.info(
@@ -291,9 +224,8 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
         )
         return job["jobId"]
 
-    def _prepare_container_properties(self, logger, command):
+    def _prepare_container_properties(self, logger):
         image_tag = self.runtime_values.get("docker", "ubuntu:20.04")
-        volumes, mount_points = self._prepare_mounts(logger, command)
         vcpu = self.runtime_values.get("cpu", 1)
         memory_mbytes = max(
             math.ceil(self.runtime_values.get("memory_reservation", 0) / 1048576), 1024
@@ -303,7 +235,7 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
             "exit_code=0",
             "bash -l ../command >> ../stdout.txt 2> >(tee -a ../stderr.txt >&2) || exit_code=$?",
         ]
-        if self.cfg.get_bool("aws", "container_sync"):
+        if self.cfg.get_bool("aws", "container_sync", False):
             commands.append("find . -type f | xargs sync")
             commands.append("sync ../stdout.txt ../stderr.txt")
         commands.append("exit $exit_code")
@@ -318,8 +250,6 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
 
         container_properties = {
             "image": image_tag,
-            "volumes": volumes,
-            "mountPoints": mount_points,
             "command": ["/bin/bash", "-ec", "\n".join(commands)],
             "environment": [
                 {"name": ev_name, "value": ev_value}
@@ -327,14 +257,11 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
             ],
             "resourceRequirements": resource_requirements,
             "privileged": self.runtime_values.get("privileged", False),
+            "mountPoints": [{"containerPath": self._fs_mount, "sourceVolume": "file_io_root"}],
         }
 
         if self.cfg["task_runtime"].get_bool("as_user"):
-            user = (
-                f"{self._studio_efs_uid}:{self._studio_efs_uid}"
-                if self._studio_efs_uid is not None
-                else f"{os.geteuid()}:{os.getegid()}"
-            )
+            user = f"{os.geteuid()}:{os.getegid()}"
             if user.startswith("0:"):
                 logger.warning(
                     "container command will run explicitly as root, since you are root and set --as-me"
@@ -343,55 +270,6 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
 
         return container_properties
 
-    def _prepare_mounts(self, logger, command):
-        """
-        Prepare the "volumes" and "mountPoints" for the Batch job definition, assembling the
-        in-container filesystem with the shared working directory, read-only input files, and
-        command/stdout/stderr files.
-        """
-
-        # prepare control files
-        with open(os.path.join(self.host_dir, "command"), "w") as outfile:
-            outfile.write(command)
-        with open(self.host_stdout_txt(), "w"):
-            pass
-        with open(self.host_stderr_txt(), "w"):
-            pass
-
-        # EFS mount point
-        volumes = [
-            {
-                "name": "efs",
-                "efsVolumeConfiguration": {
-                    "fileSystemId": self._fs_id,
-                    "transitEncryption": "ENABLED",
-                },
-            }
-        ]
-        if self._fsap_id:
-            volumes[0]["efsVolumeConfiguration"]["authorizationConfig"] = {
-                "accessPointId": self._fsap_id
-            }
-        mount_points = [{"containerPath": self._fs_mount, "sourceVolume": "efs"}]
-
-        if self._inputs_copied:
-            return volumes, mount_points
-
-        # Prepare symlinks to the input Files & Directories
-        container_prefix = os.path.join(self.container_dir, "work/_miniwdl_inputs/")
-        link_dirs_made = set()
-        for host_fn, container_fn in self.input_path_map.items():
-            assert container_fn.startswith(container_prefix) and len(container_fn) > len(
-                container_prefix
-            )
-            link_dn = os.path.dirname(container_fn)
-            if link_dn not in link_dirs_made:
-                os.makedirs(link_dn)
-                link_dirs_made.add(link_dn)
-            symlink_force(host_fn, container_fn)
-
-        return volumes, mount_points
-
     def _cleanup_job_definition(self, logger, cleanup, aws_batch, job_def_handle):
         def deregister(logger, aws_batch, job_def_handle):
             try:
@@ -413,7 +291,7 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
         """
         Poll for Batch job success or failure & return exit code
         """
-        describe_period = self.cfg.get_float("aws", "describe_period")
+        describe_period = self.cfg.get_float("aws", "describe_period", 1.0)
         cleanup.callback((lambda job_id: self._describer.unsubscribe(job_id)), job_id)
         poll_stderr = cleanup.enter_context(self.poll_stderr_context(logger))
         exit_code = None
@@ -490,14 +368,162 @@ class BatchJob(WDL.runtime.task_container.TaskContainer):
 
     def _submit_period_multiplier(self):
         if self._describer.jobs:
-            b = self.cfg.get_float("aws", "submit_period_b")
+            b = self.cfg.get_float("aws", "submit_period_b", 0.0)
             if b > 0.0:
                 t = time.time() - self._init_time
-                c = self.cfg.get_float("aws", "submit_period_c")
+                c = self.cfg.get_float("aws", "submit_period_c", 0.0)
                 return max(1.0, c - t / b)
         return 1.0
 
 
+class BatchJob(BatchJobBase):
+    """
+    EFS-based implementation, including the case of SageMaker Studio's built-in EFS. Assumes we're
+    running on an EC2 instance or Fargate container mounting an EFS Access Point at [file_io] root,
+    and configures each Batch job with the same mount.
+    """
+
+    @classmethod
+    def global_init(cls, cfg, logger):
+        super().global_init(cfg, logger)
+
+        # EFS configuration based on:
+        # - [aws] fsap / MINIWDL__AWS__FSAP
+        # - [aws] fs / MINIWDL__AWS__FS
+        # - SageMaker Studio metadata, if applicable
+        cls._fs_id = None
+        cls._fsap_id = None
+        if cfg.has_option("aws", "fs"):
+            cls._fs_id = cfg.get("aws", "fs")
+        if cfg.has_option("aws", "fsap"):
+            cls._fsap_id = cfg.get("aws", "fsap")
+            if not cls._fs_id:
+                cls._fs_id = efs_id_from_access_point(cls._region_name, cls._fsap_id)
+        cls._studio_efs_uid = None
+        sagemaker_studio_efs = detect_sagemaker_studio_efs(logger, region_name=cls._region_name)
+        if sagemaker_studio_efs:
+            (
+                studio_efs_id,
+                studio_efs_uid,
+                studio_efs_home,
+                studio_efs_mount,
+            ) = sagemaker_studio_efs
+            assert (
+                not cls._fs_id or cls._fs_id == studio_efs_id
+            ), "Configured EFS ([aws] fs / MINIWDL__AWS__FS, [aws] fsap / MINIWDL__AWS__FSAP) isn't associated with current SageMaker Studio domain EFS"
+            cls._fs_id = studio_efs_id
+            assert (
+                cls._fs_mount.rstrip("/") == studio_efs_mount.rstrip("/")
+            ) or cls._fs_mount.startswith(studio_efs_mount.rstrip("/") + "/"), (
+                "misconfiguration, set [file_io] root / MINIWDL__FILE_IO__ROOT to "
+                + studio_efs_mount.rstrip("/")
+            )
+            cls._studio_efs_uid = studio_efs_uid
+            if not cls._fsap_id:
+                cls._fsap_id = detect_studio_fsap(
+                    logger,
+                    studio_efs_id,
+                    studio_efs_uid,
+                    studio_efs_home,
+                    region_name=cls._region_name,
+                )
+                assert (
+                    cls._fsap_id
+                ), "Unable to detect suitable EFS Access Point for use with SageMaker Studio; set [aws] fsap / MINIWDL__AWS__FSAP"
+            # TODO: else sanity-check that FSAP's root directory equals studio_efs_home
+        assert (
+            cls._fs_id
+        ), "Missing EFS configuration ([aws] fs / MINIWDL__AWS__FS or [aws] fsap / MINIWDL__AWS__FSAP)"
+        if not cls._fsap_id:
+            logger.warning(
+                "AWS BatchJob plugin recommends using EFS Access Point to simplify permissions between containers (configure [aws] fsap / MINIWDL__AWS__FSAP to fsap-xxxx)"
+            )
+
+        # if no task queue in config file, try detecting miniwdl-aws-studio
+        if not cls._job_queue and sagemaker_studio_efs:
+            cls._job_queue = detect_gwfcore_batch_queue(
+                logger, sagemaker_studio_efs[0], region_name=cls._region_name
+            )
+        assert (
+            cls._job_queue
+        ), "Missing AWS Batch job queue configuration ([aws] task_queue / MINIWDL__AWS__TASK_QUEUE)"
+
+        logger.info(
+            _(
+                "initialized AWS BatchJob (EFS) plugin",
+                region_name=cls._region_name,
+                job_queue=cls._job_queue,
+                resource_limits=cls._resource_limits,
+                file_io_root=cls._fs_mount,
+                efs_id=cls._fs_id,
+                efsap_id=cls._fsap_id,
+            )
+        )
+
+    def _prepare_container_properties(self, logger):
+        container_properties = super()._prepare_container_properties(logger)
+
+        # add EFS volume & mount point
+        volumes = [
+            {
+                "name": "file_io_root",
+                "efsVolumeConfiguration": {
+                    "fileSystemId": self._fs_id,
+                    "transitEncryption": "ENABLED",
+                },
+            }
+        ]
+        if self._fsap_id:
+            volumes[0]["efsVolumeConfiguration"]["authorizationConfig"] = {
+                "accessPointId": self._fsap_id
+            }
+        container_properties["volumes"] = volumes
+
+        # set Studio UID if appropriate
+        if self.cfg["task_runtime"].get_bool("as_user") and self._studio_efs_uid:
+            container_properties["user"] = f"{self._studio_efs_uid}:{self._studio_efs_uid}"
+
+        return container_properties
+
+
+class BatchJobNoEFS(BatchJobBase):
+    """
+    Implementation assuming the Batch compute environment is configured to mount the shared
+    filesystem without further specification by us; e.g. FSxL mounted by cloud-init user data
+    script.
+    """
+
+    @classmethod
+    def global_init(cls, cfg, logger):
+        super().global_init(cfg, logger)
+
+        assert (
+            cls._job_queue
+        ), "Missing AWS Batch job queue configuration ([aws] task_queue / MINIWDL__AWS__TASK_QUEUE)"
+
+        logger.info(
+            _(
+                "initialized AWS BatchJob plugin",
+                region_name=cls._region_name,
+                job_queue=cls._job_queue,
+                resource_limits=cls._resource_limits,
+                file_io_root=cls._fs_mount,
+            )
+        )
+
+    def _prepare_container_properties(self, logger):
+        container_properties = super()._prepare_container_properties(logger)
+
+        container_properties["volumes"] = [
+            {
+                "name": "file_io_root",
+                "host": {"sourcePath": self._fs_mount},
+            }
+        ]
+
+        return container_properties
+
+
 class BatchJobDescriber:
     """
     This singleton object handles calling the AWS Batch DescribeJobs API with up to 100 job IDs
