@@ -20,7 +20,9 @@ from ._util import detect_aws_region, randomize_job_name, END_OF_LOG, efs_id_fro
 def miniwdl_submit_awsbatch(argv):
     # Configure from arguments/environment/tags
     args, unused_args = parse_args(argv)
-    verbose = args.follow or "--verbose" in unused_args or "--debug" in unused_args
+    verbose = (
+        args.follow or args.self_test or "--verbose" in unused_args or "--debug" in unused_args
+    )
     detect_env_args(args)
     if verbose:
         print("Workflow job queue: " + args.workflow_queue, file=sys.stderr)
@@ -36,96 +38,50 @@ def miniwdl_submit_awsbatch(argv):
     detect_tags_args(aws_batch, args)
 
     if verbose:
-        print("Workflow IAM role ARN: " + args.workflow_role, file=sys.stderr)
         print("Task job queue: " + args.task_queue, file=sys.stderr)
-        print("EFS Access Point: " + args.fsap, file=sys.stderr)
+        if args.efs:
+            print("Workflow IAM role ARN: " + args.workflow_role, file=sys.stderr)
+            print("EFS Access Point: " + args.fsap, file=sys.stderr)
 
-    fs_id = efs_id_from_access_point(aws_region_name, args.fsap)
-    if verbose:
-        print("EFS: " + fs_id, file=sys.stderr)
+    fs_id = None
+    if args.efs:
+        fs_id = efs_id_from_access_point(aws_region_name, args.fsap)
+        if verbose:
+            print("EFS: " + fs_id, file=sys.stderr)
 
     # TODO: accept local WDL source code; check, zip, & attach it
 
     # Prepare workflow job: command, environment, and container properties
-    job_name, miniwdl_run_cmd = form_miniwdl_run_cmd(args, unused_args)
+    job_name, miniwdl_run_cmd, wdl_zip = form_miniwdl_run_cmd(args, unused_args, verbose)
     job_name = randomize_job_name(job_name)
-
-    environment = [
-        {"name": "MINIWDL__AWS__FS", "value": fs_id},
-        {"name": "MINIWDL__AWS__FSAP", "value": args.fsap},
-        {"name": "MINIWDL__AWS__TASK_QUEUE", "value": args.task_queue},
-        {"name": "MINIWDL__FILE_IO__ROOT", "value": args.mount},
-    ]
-    extra_env = set()
-    if not args.no_env:
-        # pass through environment variables starting with MINIWDL__ (except those specific to
-        # workflow job launch, or passed through via command line)
-        for k in os.environ:
-            if k.startswith("MINIWDL__") and k not in (
-                "MINIWDL__AWS__FS",
-                "MINIWDL__AWS__FSAP",
-                "MINIWDL__AWS__TASK_QUEUE",
-                "MINIWDL__AWS__WORKFLOW_QUEUE",
-                "MINIWDL__AWS__WORKFLOW_ROLE",
-                "MINIWDL__AWS__WORKFLOW_IMAGE",
-                "MINIWDL__AWS__S3_UPLOAD_FOLDER",
-                "MINIWDL__AWS__S3_UPLOAD_DELETE_AFTER",
-                "MINIWDL__FILE_IO__ROOT",
-            ):
-                environment.append({"name": k, "value": os.environ[k]})
-                extra_env.add(k)
-
     if verbose:
         print("Workflow job image: " + args.image, file=sys.stderr)
-        if extra_env:
-            print(
-                "Passing through environment variables (--no-env to disable): "
-                + " ".join(list(extra_env)),
-                file=sys.stderr,
-            )
         print("Invocation: " + " ".join(shlex.quote(s) for s in miniwdl_run_cmd), file=sys.stderr)
-
-    workflow_container_props = {
-        "fargatePlatformConfiguration": {"platformVersion": "1.4.0"},
-        "executionRoleArn": args.workflow_role,
-        "jobRoleArn": args.workflow_role,
-        "resourceRequirements": [
-            {"type": "VCPU", "value": str(args.cpu)},
-            {"type": "MEMORY", "value": str(args.memory_GiB * 1024)},
-        ],
-        "volumes": [
-            {
-                "name": "efs",
-                "efsVolumeConfiguration": {
-                    "fileSystemId": fs_id,
-                    "transitEncryption": "ENABLED",
-                    "authorizationConfig": {"accessPointId": args.fsap},
-                },
-            }
-        ],
-        "mountPoints": [{"containerPath": args.mount, "sourceVolume": "efs"}],
-        "image": args.image,
-        "command": miniwdl_run_cmd,
-        "environment": environment,
-    }
-    if not args.no_public_ip:
-        workflow_container_props["networkConfiguration"] = {"assignPublicIp": "ENABLED"}
+    workflow_container_props = form_workflow_container_props(args, miniwdl_run_cmd, fs_id, verbose)
 
     # Register & submit workflow job
     workflow_job_def = aws_batch.register_job_definition(
         jobDefinitionName=job_name,
-        platformCapabilities=["FARGATE"],
+        platformCapabilities=["FARGATE" if args.efs else "EC2"],
         type="container",
         containerProperties=workflow_container_props,
     )
     workflow_job_def_handle = (
         f"{workflow_job_def['jobDefinitionName']}:{workflow_job_def['revision']}"
     )
+    environment2 = []
+    if wdl_zip:
+        # if the command line provided a local WDL (or WDL zipped by `miniwdl zip`), ship it in
+        # the workflow job environment, to be picked up by miniwdl-run-s3upload.
+        # we send this in the SubmitJob containerOverrides rather than RegisterJobDefinition to
+        # minimize the chance of exceeding the payload size limit for either request.
+        environment2.append({"name": "WDL_ZIP", "value": wdl_zip})
     try:
         workflow_job_id = aws_batch.submit_job(
             jobName=job_name,
             jobQueue=args.workflow_queue,
             jobDefinition=workflow_job_def_handle,
+            containerOverrides={"environment": environment2},
         )["jobId"]
         if verbose:
             print(f"Submitted {job_name} to {args.workflow_queue}:", file=sys.stderr)
@@ -176,7 +132,16 @@ def parse_args(argv):
         " or detect from DefaultFsap tag on workflow job queue]",
     )
     group.add_argument(
-        "--mount", default="/mnt/efs", help="EFS mount point in all containers [/mnt/efs]"
+        "--no-efs",
+        "--no-EFS",
+        action="store_false",
+        dest="efs",
+        help="instead of EFS, expect EC2 compute environments to automatically mount some other shared filesystem [env MINIWDL__AWS__FS=0]",
+    )
+    group.add_argument(
+        "--mount",
+        default=None,
+        help="shared filesystem mount point in all containers [/mnt/efs or /mnt/net]",
     )
     group = parser.add_argument_group("Workflow job provisioning")
     group.add_argument(
@@ -207,16 +172,16 @@ def parse_args(argv):
     group.add_argument(
         "--dir",
         default=None,
-        help="Run directory prefix [/mnt/efs/miniwdl_run]",
+        help="run directory prefix [{mount}/miniwdl_run or {mount}/miniwdl_run]",
     )
     group.add_argument(
         "--s3upload",
-        help="s3://bucket/folder/ at which to upload run outputs (otherwise left on EFS)",
+        help="s3://bucket/folder/ at which to upload run outputs (otherwise left on shared filesystem)",
     )
     group.add_argument(
         "--delete-after",
         choices=("always", "success", "failure"),
-        help="with --s3upload, delete EFS run directory afterwards",
+        help="with --s3upload, delete run directory afterwards",
     )
     parser.add_argument(
         "--wait", "-w", action="store_true", help="wait for workflow job to complete"
@@ -231,6 +196,10 @@ def parse_args(argv):
 
     args, unused_args = parser.parse_known_args(argv[1:])
 
+    if os.environ.get("MINIWDL__AWS__FS", "").strip().lower() in ("false", "f", "0", "no", "n"):
+        args.efs = False
+    if not args.mount:
+        args.mount = "/mnt/efs" if args.efs else "/mnt/net"
     if args.mount.endswith("/"):
         args.mount = args.mount[:-1]
     assert args.mount
@@ -270,6 +239,7 @@ def detect_env_args(args):
     )
     args.image = args.image if args.image else os.environ.get("MINIWDL__AWS__WORKFLOW_IMAGE", None)
     if not args.image:
+        # version-matched default image from our GitHub build
         import importlib_metadata
 
         try:
@@ -282,6 +252,7 @@ def detect_env_args(args):
                 file=sys.stderr,
             )
             sys.exit(1)
+
     if args.delete_after and not args.s3upload:
         print("--delete-after requires --s3upload", file=sys.stderr)
         sys.exit(1)
@@ -298,24 +269,13 @@ def detect_env_args(args):
 def detect_tags_args(aws_batch, args):
     """
     If not otherwise set by command line arguments or environment, inspect tags of the workflow job
-    queue to detect default EFS Access Point ID (fsap-xxx), task job queue, and/or workflow role
-    ARN. Infra provisioning (CloudFormation, Terraform, etc.) may have set the respective tags.
+    queue to detect default task job queue and (if applicable) EFS Access Point ID and workflow
+    role ARN. Infra provisioning (CloudFormation, Terraform, etc.) may have set the expected tags.
     """
-    if not (args.fsap or args.task_queue or args.workflow_role):
+    if not args.task_queue or (args.efs and not (args.fsap or args.workflow_role)):
         workflow_queue_tags = aws_batch.describe_job_queues(jobQueues=[args.workflow_queue])[
             "jobQueues"
         ][0]["tags"]
-        if not args.fsap:
-            try:
-                args.fsap = workflow_queue_tags["DefaultFsap"]
-                assert args.fsap.startswith("fsap-")
-            except:
-                print(
-                    "Unable to detect default EFS Access Point (fsap-xxxx) from DefaultFsap tag of workflow job queue."
-                    " Set --fsap or environment variable MINIWDL__AWS__FSAP.",
-                    file=sys.stderr,
-                )
-                sys.exit(1)
         if not args.task_queue:
             args.task_queue = workflow_queue_tags.get("DefaultTaskQueue", None)
             if not args.task_queue:
@@ -325,7 +285,19 @@ def detect_tags_args(aws_batch, args):
                     file=sys.stderr,
                 )
                 sys.exit(1)
-        if not args.workflow_role:
+        if args.efs and not args.fsap:
+            try:
+                args.fsap = workflow_queue_tags["DefaultFsap"]
+                assert args.fsap.startswith("fsap-")
+            except:
+                if not args.fsap:
+                    print(
+                        "Unable to detect default EFS Access Point (fsap-xxxx) from DefaultFsap tag of workflow job queue."
+                        " Set --fsap or environment variable MINIWDL__AWS__FSAP.",
+                        file=sys.stderr,
+                    )
+                    sys.exit(1)
+        if args.efs and not args.workflow_role:
             # Workflow role ARN is needed for Fargate Batch (unlike EC2 Batch, where a role is
             # associated with the EC2 instance profile in the compute environment).
             try:
@@ -343,10 +315,11 @@ def detect_tags_args(aws_batch, args):
                     sys.exit(1)
 
 
-def form_miniwdl_run_cmd(args, unused_args):
+def form_miniwdl_run_cmd(args, unused_args, verbose=False):
     """
     Formulate the `miniwdl run` command line to be invoked in the workflow job container
     """
+    wdl_zip = None
     if args.self_test:
         self_test_dir = os.path.join(
             args.mount, "miniwdl_run_self_test", datetime.today().strftime("%Y%m%d_%H%M%S")
@@ -354,10 +327,17 @@ def form_miniwdl_run_cmd(args, unused_args):
         miniwdl_run_cmd = ["miniwdl", "run_self_test", "--dir", self_test_dir]
         job_name = args.name if args.name else "miniwdl_run_self_test"
     else:
-        wdl_filename = next((arg for arg in unused_args if not arg.startswith("-")), None)
-        if not wdl_filename:
+        wdl_filename_pos = next(
+            (i for i, arg in enumerate(unused_args) if not arg.startswith("-")), -1
+        )
+        if wdl_filename_pos < 0:
             print("Command line appears to be missing WDL filename", file=sys.stderr)
             sys.exit(1)
+        wdl_filename = unused_args[wdl_filename_pos]
+        wdl_zip = zip_wdl(wdl_filename, verbose)
+        if wdl_zip:
+            # this sentinel argument will be recognized by miniwdl-run-s3upload
+            unused_args[wdl_filename_pos] = "--WDL--ZIP--"
         job_name = args.name
         if not job_name:
             job_name = os.path.basename(wdl_filename).lstrip(".")
@@ -373,7 +353,149 @@ def form_miniwdl_run_cmd(args, unused_args):
         miniwdl_run_cmd.extend(["--dir", args.dir])
         miniwdl_run_cmd.extend(["--s3upload", args.s3upload] if args.s3upload else [])
         miniwdl_run_cmd.extend(["--delete-after", args.delete_after] if args.delete_after else [])
-    return (job_name, miniwdl_run_cmd)
+    return (job_name, miniwdl_run_cmd, wdl_zip)
+
+
+def zip_wdl(wdl_filename, verbose=False):
+    """
+    If wdl_filename is an existing local .wdl or .zip file, prepare to ship it as the WDL source
+    code for the workflow job to execute. (Otherwise, it'll be passed through assuming it's some
+    path or URI the workflow job will be able to open directly.)
+
+    If it's a .zip file, assume it's generated by `miniwdl zip`.
+
+    If it's a .wdl file, run `miniwdl zip` on it.
+    """
+    if not os.path.isfile(wdl_filename) or not (
+        wdl_filename.endswith(".wdl") or wdl_filename.endswith(".zip")
+    ):
+        if verbose:
+            print(
+                f"WDL: {wdl_filename} (not a local WDL file; assuming accessible inside workflow job)"
+            )
+        return None
+
+    # load zip bytes
+    if wdl_filename.endswith(".wdl"):
+        import subprocess
+        import tempfile
+
+        try:
+            with tempfile.TemporaryDirectory() as tmpdir:
+                zip_fn = os.path.join(tmpdir, os.path.basename(wdl_filename)) + ".zip"
+                subprocess.check_call(["miniwdl", "zip", "-o", zip_fn, wdl_filename])
+                with open(zip_fn, "rb") as zip_file:
+                    zip_bytes = zip_file.read()
+                # TODO: detect -i file.json in unused_args and provide it to miniwdl zip too
+        except subprocess.CalledProcessError as exn:
+            sys.exit(exn.returncode)
+    else:
+        assert wdl_filename.endswith(".zip")
+        with open(wdl_filename, "rb") as zip_file:
+            zip_bytes = zip_file.read()
+    assert zip_bytes, "empty WDL zip"
+
+    # aggressively compress, to maximize chance of fitting within the 30KiB limit on Batch
+    # SubmitJob request: https://docs.aws.amazon.com/batch/latest/userguide/service_limits.html
+    import base64
+    import lzma
+
+    zip_str = base64.b85encode(
+        lzma.compress(zip_bytes, format=lzma.FORMAT_ALONE, preset=(9 | lzma.PRESET_EXTREME))
+    ).decode("ascii")
+    if verbose:
+        print(
+            f"WDL/ZIP: {wdl_filename} (encoded as {len(zip_str)} bytes to submit with workflow job)",
+            file=sys.stderr,
+        )
+    return zip_str
+
+
+def form_workflow_container_props(args, miniwdl_run_cmd, fs_id, verbose=False):
+    environment = [
+        {"name": "MINIWDL__AWS__TASK_QUEUE", "value": args.task_queue},
+        {"name": "MINIWDL__FILE_IO__ROOT", "value": args.mount},
+    ]
+    if args.efs:
+        environment.append({"name": "MINIWDL__AWS__FS", "value": fs_id})
+        environment.append({"name": "MINIWDL__AWS__FSAP", "value": args.fsap})
+    else:
+        environment.append(
+            {"name": "MINIWDL__SCHEDULER__CONTAINER_BACKEND", "value": "aws_batch_job_no_efs"}
+        )
+    extra_env = set()
+    if not args.no_env:
+        # pass through environment variables starting with MINIWDL__ (except those specific to
+        # workflow job launch, or passed through via command line)
+        for k in os.environ:
+            if k.startswith("MINIWDL__") and k not in (
+                "MINIWDL__AWS__FS",
+                "MINIWDL__AWS__FSAP",
+                "MINIWDL__AWS__TASK_QUEUE",
+                "MINIWDL__AWS__WORKFLOW_QUEUE",
+                "MINIWDL__AWS__WORKFLOW_ROLE",
+                "MINIWDL__AWS__WORKFLOW_IMAGE",
+                "MINIWDL__AWS__S3_UPLOAD_FOLDER",
+                "MINIWDL__AWS__S3_UPLOAD_DELETE_AFTER",
+                "MINIWDL__FILE_IO__ROOT",
+            ):
+                environment.append({"name": k, "value": os.environ[k]})
+                extra_env.add(k)
+
+    if verbose and extra_env:
+        print(
+            "Passing through environment variables (--no-env to disable): "
+            + " ".join(list(extra_env)),
+            file=sys.stderr,
+        )
+
+    workflow_container_props = {
+        "resourceRequirements": [
+            {"type": "VCPU", "value": str(args.cpu)},
+            {"type": "MEMORY", "value": str(args.memory_GiB * 1024)},
+        ],
+        "image": args.image,
+        "command": miniwdl_run_cmd,
+        "environment": environment,
+    }
+    if args.efs:
+        # EFS: set EFS volume/mountPoint and Fargate execution role
+        assert args.workflow_role and fs_id and args.fsap
+        workflow_container_props.update(
+            {
+                "fargatePlatformConfiguration": {"platformVersion": "1.4.0"},
+                "executionRoleArn": args.workflow_role,
+                "jobRoleArn": args.workflow_role,
+                "volumes": [
+                    {
+                        "name": "efs",
+                        "efsVolumeConfiguration": {
+                            "fileSystemId": fs_id,
+                            "transitEncryption": "ENABLED",
+                            "authorizationConfig": {"accessPointId": args.fsap},
+                        },
+                    }
+                ],
+                "mountPoints": [{"containerPath": args.mount, "sourceVolume": "efs"}],
+            }
+        )
+        if not args.no_public_ip:
+            workflow_container_props["networkConfiguration"] = {"assignPublicIp": "ENABLED"}
+    else:
+        # non-EFS: set volume/mountPoint assuming compute environments mount automatically
+        workflow_container_props.update(
+            {
+                "volumes": [
+                    {
+                        "name": "file_io_root",
+                        "host": {"sourcePath": args.mount},
+                    }
+                ],
+                "mountPoints": [{"containerPath": args.mount, "sourceVolume": "file_io_root"}],
+            }
+        )
+
+    return workflow_container_props
 
 
 def wait(aws_region_name, aws_batch, workflow_job_id, follow, expect_log_eof=True):
